import json
import datetime
import os

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.trace import set_span_in_context

from opentelemetry import propagate
from opentelemetry.propagators.textmap import DefaultSetter
from opentelemetry.propagators.textmap import DefaultGetter

from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.sdk.trace import ReadableSpan
import typing
from pathlib import Path
from os import linesep
from typing import Optional

from rich.traceback import install
from rich.logging import RichHandler
from rich import print, inspect, print_json
from rich.panel import Panel
from rich.padding import Padding

install()


class FileSpanExporter(SpanExporter):
    """
    Implementation of :class:`SpanExporter` that prints spans to a given file.
    """

    def __init__(
        self,
        file_path_getter: typing.Callable[[], str],
        service_name: Optional[str] = None,
        formatter: typing.Callable[[ReadableSpan], str] = lambda span: span.to_json()
        + linesep,
    ):
        self.file_path_getter = file_path_getter
        self.formatter = formatter
        self.service_name = service_name
        self.buffered_spans = []

    def export(self, spans: typing.Sequence[ReadableSpan]) -> SpanExportResult:
        """
        If the path returned by self.file_path_getter exists,
        writes the given spans (and any previously saved spans)
        to a file under that path.
        Else, it saves the spans in self.buffered_spans,
        in hopes that in a future call the path will have been created.
        """
        self.buffered_spans.append(spans)

        file_path = self.file_path_getter()

        if file_path is not None and Path(file_path).exists():
            with open(Path(file_path, "otlp_trace"), "a") as file:
                for bspans in self.buffered_spans:
                    for span in bspans:
                        file.write(self.formatter(span))
                file.flush()
            self.buffered_spans.clear()

        return SpanExportResult.SUCCESS

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return True


class EventsOTLP:
    """
    Handles event tracing using OpenTelemetry
    (or the console on the user's local environment).
    """

    def __init__(self, log=None, destination=None, otlp_url=None, trace_context=None):        
        self.log = log
        self.otlp_url = otlp_url # "localhost:3417"
        self.destination = destination
        self.seq = 0

        # if the trace was started on the cluster by a different entity (e.g. the worker),
        # its context is given by an environment variable.            
        if trace_context:            
            self.otlp_context = {"traceparent": trace_context}
        else:
            self.otlp_context = None

        self.otlp_processor = None
        self.file_processor = None
        self.tracer = None

    def setup_tracer(self):
        # Service name is required for most backends
        resource = Resource(attributes={SERVICE_NAME: "citros-sim"})

        provider = TracerProvider(resource=resource)

        # the citros run dir is only created before a simulation run,
        # so it must be retrieved dynamically.
        if self.destination:
            file_exporter = FileSpanExporter(lambda: self.destination)
            file_processor = BatchSpanProcessor(file_exporter)
            provider.add_span_processor(file_processor)
            self.file_processor = file_processor

        if self.otlp_url:
            # when running in the cluster, the telop collector (k8s deamonset)
            # is allways local for the pod it is running on.
            otlp_exporter = OTLPSpanExporter(
                endpoint=self.self.otlp_url, insecure=True
            )
            otlp_processor = BatchSpanProcessor(otlp_exporter)
            provider.add_span_processor(otlp_processor)
            self.otlp_processor = otlp_processor

        trace.set_tracer_provider(provider)

        return trace.get_tracer("citros-events-tracer")

    def on_shutdown(self):
        """
        Ensures that any remaining spans in the queue are flushed and sent to the exporter.
        """
        if self.file_processor:
            self.file_processor.shutdown()
        if self.otlp_processor:
            self.otlp_processor.shutdown()

    def serialize_current_context(self):
        """
        return value is a dictionary of the form
        {'traceparent': '00-<trace_id>-<span-id>-<some other id>'}
        example:
        {'traceparent': '00-cb9cfff1dd2ba53e9b09360fedb78bb7-bcbc4396d4ae74c4-01'}
        """
        carrier = {}
        PROPAGATOR = propagate.get_global_textmap()
        PROPAGATOR.inject(carrier, setter=DefaultSetter())
        return carrier

    def deserialize_to_context(self, metadata):
        PROPAGATOR = propagate.get_global_textmap()
        return PROPAGATOR.extract(metadata, getter=DefaultGetter())

    def set_span_attributes(
        self, span, event_type, tag, message, metadata
    ):
        # span.set_attribute("batch-run-id", batch_run_id)
        # span.set_attribute("sid", sid)
        span.set_attribute("event-type", event_type)
        span.set_attribute("tag", tag)
        span.set_attribute("message", message)
        span.set_attribute("metadata", metadata)
        span.set_attribute("created", datetime.datetime.now().isoformat())
        span.set_attribute("emit-sequence", self.seq)
        self.seq = self.seq + 1


    def emit(self, event_type, tag, message, metadata):
        """
        Creates a new span and sets its attributes with the given arguments.
        If self.otlp_context has not been set, the span will be started as the root span.
        Else, the span will be started with the span in the self.otlp_context as parent.

        
        :param event: event type
        :param tag: tag - can be any string
        :param message: a message for the event
        :param metadata: some dict object containing metadata.
        """
        if self.tracer is None:
            self.tracer = self.setup_tracer()

        try:
            if isinstance(metadata, dict):
                metadata = json.dumps(metadata)
        except Exception as e:            
            if self.log:
                self.log.exception(e)
            else:
                print(e)
            metadata = None

        if metadata is None:
            metadata = ""

        if self.otlp_context is None:
            with self.tracer.start_as_current_span(event_type) as span:
                self.set_span_attributes(
                    span, event_type, tag, message, metadata
                )
                self.otlp_context = self.serialize_current_context()
        else:
            # create a new span with the current span as parent.
            context = self.deserialize_to_context(self.otlp_context)
            span = self.tracer.start_span(event_type, context=context)

            with trace.use_span(span, end_on_exit=True):
                self.set_span_attributes(
                    span, event_type, tag, message, metadata
                )
                self.otlp_context = self.serialize_current_context()

    def schedule(self, tag="", message="", metadata=None):
        """
        Sends event of type SCHEDULE to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("SCHEDULE", tag, message, metadata)

    def creating(self, tag="", message="", metadata=None):
        """
        Sends event of type CREATING to CiTROS

        
        :param tag: tag - can be any string
        :param message: a message for the event
        :param metadata: some dict object containing metadata.
        """
        self.emit("CREATING", tag, message, metadata)

    def init(self, tag="", message="", metadata=None):
        """
        Sends event of type INIT to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("INIT", tag, message, metadata)

    def starting(self, tag="", message="", metadata=None):
        """
        Sends event of type STARTING to CiTROS
        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("STARTING", tag, message, metadata)

    def running(self, tag="", message="", metadata=None):
        """
        Sends event of type RUNNING to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("RUNNING", tag, message, metadata)

    def terminating(self, tag="", message="", metadata=None):
        """
        Sends event of type TERMINATING to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("TERMINATING", tag, message, metadata)

    def stopping(self, tag="", message="", metadata=None):
        """
        Sends event of type STOPPING to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("STOPPING", tag, message, metadata)

    def done(self, tag="", message="", metadata=None):
        """
        Sends event of type DONE to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("DONE", tag, message, metadata)

    def error(self, tag="", message="", metadata=None):
        """
        Sends event of type ERROR to CiTROS

        
        :param tag:  (Default value = "")
        :param message:  (Default value = "")
        :param metadata:  (Default value = None)
        """
        self.emit("ERROR", tag, message, metadata)



import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install

install()

from bin import __version__ as citros_version

from .parsers import (
    parser_run,
    parser_init,
    parser_data,
    parser_service,
    parser_report,
    parser_launch,
    parser_parameter,
    parser_simulation,
)


# PANNEL = ""
PANNEL = Panel.fit(
    f"""[green]████████████████████████████████████████████████████████████
██      ███        ██        ██       ████      ████      ██
█  ████  █████  ████████  █████  ████  ██  ████  ██  ███████
█  ███████████  ████████  █████       ███  ████  ███      ██
█  ████  █████  ████████  █████  ███  ███  ████  ████████  █
██      ███        █████  █████  ████  ███      ████      ██
████████████████████████████████████████████████████████████""",
    subtitle=f"[{citros_version}]",
)


def epilog(url="https://docs.citros.io"):
    return Panel(Markdown(f"Read more at `{url}`"))


def main():
    description_path = "citros.md"

    # main parser -----------------------------------------
    parser = argparse.ArgumentParser(
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog(),
        formatter_class=RichHelpFormatter,
        # add_help=False
    )

    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.add_argument("-V", "--version", action="version", version=citros_version)
    parser.set_defaults(func=citros)

    def func(args, argv):
        print(PANNEL)
        citros(args, argv)

    parser.set_defaults(func=func)

    subparsers = parser.add_subparsers(
        title="commands",
        help="citros commands",
        dest="type",  # required=True
    )

    parser_init(
        subparsers,
        epilog("https://docs.citros.io/docs/cli_commands#command-init"),
    )
    parser_run(
        subparsers,
        epilog("https://docs.citros.io/docs/cli_commands#command-run"),
    )
    parser_data(
        subparsers,
        epilog(),  # TODO: fill with correct doc URL
    )
    parser_report(
        subparsers,
        epilog(),  # TODO: fill with correct doc URL
    )
    parser_service(
        subparsers,
        epilog(),  # TODO: fill with correct doc URL
    )

    # parser_simulation(
    #     subparsers,
    #     epilog(),  # TODO: fill with correct doc URL
    # )
    # parser_parameter(
    #     subparsers,
    #     epilog(),  # TODO: fill with correct doc URL
    # )
    # parser_launch(
    #     subparsers,
    #     epilog(),  # TODO: fill with correct doc URL
    # )

    args, argv = parser.parse_known_args()

    args.func(args, argv)


if __name__ == "__main__":
    main()



import os
import glob
import json
import importlib_resources

from bin import __version__ as citros_version
from time import sleep
from citros import Citros
from pathlib import Path
from rich import box, print, inspect, print_json, pretty
from rich.table import Table
from rich.console import Console
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter


pretty.install()

from InquirerPy import prompt, inquirer
from rich.prompt import Prompt
from InquirerPy.base.control import Choice
from InquirerPy.separator import Separator
from prompt_toolkit.validation import Validator, ValidationError

from citros import (
    Batch,
    CitrosNotFoundException,
    str_to_bool,
    suppress_ros_lan_traffic,
    Report,
    NoNotebookFoundException,
    NoConnectionToCITROSDBException,
)
from .config import config


class NumberValidator(Validator):
    """
    small helper class for validating user input during an interactive session.
    """

    def validate(self, document):
        try:
            int(document.text)
        except ValueError:
            raise ValidationError(
                message="Please enter a number", cursor_position=len(document.text)
            )


keybindings = {
    "skip": [{"key": "c-c"}, {"key": "escape"}],
}


def exit_citros_cli():
    print("[green]Bye👋")
    exit(0)


def citros(args, argv):
    # action
    action = inquirer.select(
        raise_keyboard_interrupt=False,
        mandatory=False,
        keybindings=keybindings,
        message="Select Action:",
        choices=[
            Choice("init", name="🍋 Init: initialize .citros in current directory"),
            Choice("run", name="🔥 Run: new simulation"),
            Choice("data", name="📊 Data: for data managment "),
            Choice("report", name="📝 Report: generation and management"),
            Choice("service", name="🔖 Service: CITROS API service functions"),
            Separator(),
            Choice("exit", name="EXIT"),
        ],
        default="",
        border=True,
    ).execute()

    if action == "data":
        data(args, argv)
    elif action == "report":
        report(args, argv)
    elif action == "run":
        run(args, argv)
    elif action == "init":
        init(args, argv)
    elif action == "service":
        service(args, argv)
    elif action == "exit":
        exit_citros_cli()
    elif action is None:
        exit_citros_cli()
    else:
        print("[red]Error: unknown action")


############################# CLI implementation ##############################
def init(args, argv):
    """
    :param args.dir:
    :param args.debug:
    :param args.verbose:
    """
    print(f'initializing CITROS at "{Path(args.dir).resolve()}". ')
    citros = Citros(new=True, root=args.dir, verbose=args.verbose, debug=args.debug)
    if args.debug:
        print("[green]done initializing CITROS")


def run(args, argv):
    """
    :param args.simulation_name:
    :param args.index:
    :param args.completions:

    :param args.batch_name:
    :param args.batch_message:

    :param args.lan_traffic:

    :param args.debug:
    :param args.verbose:
    """
    # inspect(args)
    is_interactive = False
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    if args.debug:
        print("[green]done initializing CITROS")

    if not hasattr(args, "batch_name") or args.batch_name is None:
        is_interactive = True
        batch_name = Prompt.ask("Please name this batch run", default="citros")
    else:
        batch_name = args.batch_name

    if not hasattr(args, "batch_message") or args.batch_message is None:
        batch_message = Prompt.ask(
            "Enter a message for this batch run",
            default="This is a default batch message from citros",
        )
    else:
        batch_message = args.batch_message

    if is_interactive or not hasattr(args, "completions"):
        completions = Prompt.ask(
            "How many times you want the simulation to run?",
            default="1",
        )
    else:
        completions = args.completions

    if not batch_name and str_to_bool(citros.settings["force_batch_name"]):
        print("[red]Please supply a batch name with flag -n <name>.")
        print(
            Panel.fit(
                Padding('You may run [green]"citros run -n <name>" ', 1), title="help"
            )
        )
        return False

    if not batch_message and str_to_bool(citros.settings["force_message"]):
        print("[red]Please supply a batch message with flag -m <message>.")
        print(
            Panel.fit(
                Padding('You may run [green]"citros run -m <message>"', 1), title="help"
            )
        )
        return False

    simulation_name = getattr(args, "simulation_name", None)
    simulation = choose_simulation(
        citros,
        simulation_name,
    )
    root_rec_dir = f"{args.dir}/.citros/data"
    if config.RECORDINGS_DIR:
        root_rec_dir = config.RECORDINGS_DIR

    batch_version = getattr(args, "version", None)
    batch_index = getattr(args, "index", -1)

    console = Console()
    console.rule(f"command")
    print(
        f'[white]citros run --dir {args.dir} --batch_name {batch_name} --batch_message "{batch_message}" --simulation_name {simulation_name} {"--version " + batch_version if batch_version is not None else ""} --completions {completions} --index {batch_index}'
    )
    console.rule(f"")

    batch = Batch(
        root_rec_dir,
        simulation,
        name=batch_name,
        message=batch_message,
        version=batch_version,
        verbose=args.verbose,
        debug=args.debug,
    )
    try:
        batch.run(
            completions,
            batch_index,
            ros_domain_id=config.ROS_DOMAIN_ID,
            trace_context=config.TRACE_CONTEXT,
        )
    except ModuleNotFoundError:
        print("[red]Error:[/red] ROS2 is not installed or not in your PATH.")
        print(
            Panel(
                Padding(
                    """Please install ROS2 on the system and source it:
[green]source /opt/ros/{ros2 distribution}/setup.bash[/green]

Please build your ROS2 workspace and source it by:
[green]colcon build
source install/local_setup.bash""",
                    1,
                ),
                title="help",
            )
        )
        return
    # when running multiple runs, load to DB after all runs is done.
    # if index != -1 then we run only a part of the batch, so we don't want to load to DB yet.
    if getattr(args, "index", -1) == -1:
        try:
            print("Upoading data po DB...")
            batch.unload()
            batch.upload()
        except NoConnectionToCITROSDBException:
            print("[red]CITROS DB is not running.")
            print(
                Panel.fit(
                    Padding("[green]citros data db create", 1),
                    title="help",
                )
            )
        except Exception as e:
            citros.logger.error(e)

    print(f"[green]CITROS run completed successfully. ")
    print(
        f"[green]You may run [blue]'citros data service'[/blue] to get access to your data using CITROS API."
    )


# helper function
def choose_simulation(citros: Citros, simulation_name=None):
    """
    Choose a simulation from the available simulations in the Citros object.

    Args:
        citros (Citros): The Citros object containing the simulations.
        simulation_name (str, optional): The name of the simulation to choose. Defaults to None.

    Returns:
        Simulation: The chosen simulation object.

    Raises:
        KeyError: If the specified simulation name is not found in the available simulations.
    """

    simulations_dict = {}
    for s in citros.simulations:
        simulations_dict[s.name] = s

    if simulation_name:
        return simulations_dict[simulation_name]
    sim_names = simulations_dict.keys()

    # sanity check - should never happen because internal_sync will fail if there
    #                isn't at least one simulation file.
    if not sim_names:
        print(
            f"[red]There are currently no simulations in your {citros.SIMS_DIR} folder. \
                	 Please create at least one simulation for your project."
        )
        return

    # interactive
    answers = prompt(
        [
            {
                "type": "list",
                "name": "sim_names",
                "message": "Please choose the simulation you wish to run:",
                "choices": sim_names,
            }
        ]
    )

    sim_name = answers.get("sim_names")
    return simulations_dict[sim_name]


############################# Simulation implementation ##############################
# TODO[critical]: implement
def simulation_list(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


# TODO[critical]: implement
def simulation_run(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


####################### parameter setup implementation ##############################
# TODO[critical]: implement
def parameter_setup_new(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


# TODO[critical]: implement
def parameter_setup_list(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


# TODO[critical]: implement
def parameter_setup(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


############################# DATA implementation ##############################
def data(args, argv):
    action = inquirer.select(
        raise_keyboard_interrupt=False,
        mandatory=False,
        keybindings=keybindings,
        message="Select Action:",
        choices=[
            Choice("tree", name="🌲 Tree: tree view of data"),
            Choice("list", name="*️⃣  List: reports list"),
            Choice("db", name="📂 DB: section"),
            Choice("service", name="🗳  Service: section"),
            Separator(),
            Choice("exit", name="EXIT"),
        ],
        default="",
        border=True,
    ).execute()

    if action is None:
        exit_citros_cli()

    if action == "tree":
        data_tree(args, argv)
    elif action == "list":
        data_list(args, argv)
    elif action == "db":
        data_db(args, argv)
    elif action == "service":
        service(args, argv)
    elif action == "exit":
        exit_citros_cli()
    else:
        print("[red]Error: unknown action")
    return


def data_tree(args, argv):
    """
    choose a batch run then choose an action to perform on the batch run.

    Args:
        args: The command-line arguments passed to the function.
        argv: The list of command-line arguments.

    Returns:
        None
    """

    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    chosen_simulation, chosen_batch, version = choose_batch(
        citros.root_citros,
        simulation_name=getattr(args, "simulation", None),
        batch_name=getattr(args, "batch", None),
        version=getattr(args, "version", None),
    )

    setattr(args, "simulation", chosen_simulation)
    setattr(args, "batch", chosen_batch)
    setattr(args, "version", version)

    # action
    action = inquirer.select(
        raise_keyboard_interrupt=False,
        mandatory=False,
        keybindings=keybindings,
        message="Select Action:",
        choices=[
            Choice("info", name="Info"),
            Choice("load", name="Load"),
            Choice("unload", name="Unload"),
            Choice("delete", name="Delete"),
            Separator(),
            Choice("exit", name="EXIT"),
        ],
        default="",
        border=True,
    ).execute()

    if action is None:
        exit_citros_cli()

    if action == "info":
        data_info(args, argv)
    elif action == "load":
        data_load(args, argv)
    elif action == "unload":
        data_unload(args, argv)
    elif action == "delete":
        data_delete(args, argv)
    elif action == "exit":
        exit_citros_cli()
    else:
        print("[red]Error: unknown action")


def data_list(args, argv):
    """
    List simulation runs and their details.

    Returns:
        None
    """

    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
        flat_batches = citros.get_batches_flat()
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))

        return

    table = Table(
        title=f"Simulation Runs in: [blue]{citros.root_citros / 'data'}", box=box.SQUARE
    )

    table.add_column("Simulation", style="cyan", no_wrap=True)
    table.add_column("Run name", style="magenta", justify="left")
    table.add_column("Versions", justify="left", style="green")
    table.add_column("message", style="magenta", justify="left")
    table.add_column("status", justify="right", style="green")
    table.add_column("completions", style="magenta", justify="center")
    table.add_column(
        "path", style="cyan", justify="left", no_wrap=False, overflow="fold"
    )

    for flat_batch in flat_batches:
        if flat_batch["status"] == "LOADED":
            status_clore = "green"
        elif flat_batch["status"] == "UNLOADED":
            status_clore = "yellow"
        else:
            status_clore = "red"

        table.add_row(
            flat_batch["simulation"],
            flat_batch["name"],
            flat_batch["version"],
            flat_batch["message"],
            f"[{status_clore}]{flat_batch['status']}",
            flat_batch["completions"],
            str(flat_batch["path"]).removeprefix(os.getcwd()).removeprefix("/"),
        )

    console = Console()
    console.print(table)


def data_info(args, argv):
    """
    Handle the 'info' command for Citros data.

    :param args.simulation: simulation name
    :param args.batch: batch name
    :param args.version: batch_version

    :param args.debug: Flag to indicate debug mode.
    :param args.verbose: Flag to indicate verbose console prints.
    """
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    chosen_simulation, chosen_batch, version = choose_batch(
        citros.root_citros,
        simulation_name=getattr(args, "simulation", None),
        batch_name=getattr(args, "batch", None),
        version=getattr(args, "version", None),
    )

    batch = citros.get_batch(
        simulation=chosen_simulation, name=chosen_batch, version=version
    )

    console = Console()
    console.rule(f".citros/data/{chosen_simulation}/{chosen_batch}/{version}/info.json")
    console.print_json(data=batch.data)
    console.rule(f"")


def data_load(args, argv):
    """
    Handle the 'load' command for Citros data. loads tha batch to citros postgres DB instance.

    :param args.simulation: simulation name
    :param args.batch: batch name
    :param args.version: batch_version

    :param args.debug: Flag to indicate debug mode.
    :param args.verbose: Flag to indicate verbose console prints.
    """
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    chosen_simulation, chosen_batch, version = choose_batch(
        citros.root_citros,
        simulation_name=getattr(args, "simulation", None),
        batch_name=getattr(args, "batch", None),
        version=getattr(args, "version", None),
    )
    print(f"Uploading data to DB... { chosen_simulation} / {chosen_batch} / {version}")
    batch = citros.get_batch(
        simulation=chosen_simulation, name=chosen_batch, version=version
    )

    try:
        batch.unload()
        batch.upload()
    except NoConnectionToCITROSDBException:
        print("[red]CITROS DB is not running.")
        print(
            Panel.fit(
                Padding(
                    'You may run [green]"citros data db create"[/green]  to create a new DB',
                    1,
                )
            )
        )
        return

    console = Console()
    console.rule(f"{chosen_simulation} / {chosen_batch} / {version}")
    console.print_json(data=batch.data)


def data_unload(args, argv):
    """
    Handle the 'unload' command for Citros data. unloads tha batch to citros postgres DB instance.

    :param args.simulation: simulation name
    :param args.batch: batch name
    :param args.version: batch_version

    :param args.debug: Flag to indicate debug mode.
    :param args.verbose: Flag to indicate verbose console prints.
    """
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    chosen_simulation, chosen_batch, version = choose_batch(
        citros.root_citros,
        simulation_name=getattr(args, "simulation", None),
        batch_name=getattr(args, "batch", None),
        version=getattr(args, "version", None),
    )
    print(
        f"Dropping data from DB... { chosen_simulation } / {chosen_batch} / {version}"
    )
    batch = citros.get_batch(
        simulation=chosen_simulation, name=chosen_batch, version=version
    )

    batch.unload()


def data_delete(args, argv):
    """
    Handle the 'delete' command for Citros data. delete batch from filesystem and DB.

    :param args.simulation: simulation name
    :param args.batch: batch name
    :param args.version: batch_version

    :param args.debug: Flag to indicate debug mode.
    :param args.verbose: Flag to indicate verbose console prints.
    """
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    chosen_simulation, chosen_batch, version = choose_batch(
        citros.root_citros,
        simulation_name=getattr(args, "simulation", None),
        batch_name=getattr(args, "batch", None),
        version=getattr(args, "version", None),
    )
    print(f"deleting data from { chosen_simulation } / {chosen_batch} / {version}")
    citros.delete_batch(
        simulation=chosen_simulation, name=chosen_batch, version=version
    )


def data_db(args, argv):
    action = inquirer.select(
        raise_keyboard_interrupt=False,
        mandatory=False,
        keybindings=keybindings,
        message="Select Action:",
        choices=[
            Choice(
                "create", name="▶ Create: create citros db instance and initializes it."
            ),
            Choice("remove", name="🗑  Remove: remove the db instance."),
            Choice("init", name="✏  Init: initialize the db instance"),
            Choice("logs", name="📝 Logs: show logs of DB instance"),
            Choice("status", name="❓ Status: Show weather the serivce is up or not"),
            Choice("stop", name="⏹  Stop: stops the citros db instance if running."),
            Separator(),
            Choice("exit", name="EXIT"),
        ],
        default="",
        border=True,
    ).execute()

    if action is None:
        exit_citros_cli()

    if action == "create":
        data_db_create(args, argv)
    elif action == "remove":
        data_db_remove(args, argv)
    elif action == "init":
        data_db_init(args, argv)
    elif action == "logs":
        data_db_logs(args, argv)
    elif action == "status":
        data_db_status(args, argv)
    elif action == "stop":
        data_db_stop(args, argv)
    elif action == "exit":
        exit_citros_cli()
    else:
        print("[red]Error: unknown action")


def _init_db(verbose, debug):
    """
    initializing the DB
    """
    from citros import CitrosDB

    citrosDB = CitrosDB(
        config.POSTGRES_USERNAME,
        config.POSTGRES_PASSWORD,
        config.CITROS_DATA_HOST,
        config.CITROS_DATA_PORT,
        config.POSTGRES_DATABASE,
        verbose=verbose,
        debug=debug,
    )

    citrosDB.init_db()


def data_db_create(args, argv):
    import docker

    # inspect(config)
    try:
        client = docker.from_env()
    except Exception as e:
        print(
            "[red]Docker is not running. Please start docker and try again. exiting..."
        )
        if args.verbose:
            raise e
        return

    try:
        container = client.containers.get(config.DB_CONTAINER_NAME)
        container.start()
        # inspect(container)
        print(f"[green]CITROS DB is created")
        return
    except docker.errors.NotFound:
        container = None

    container = client.containers.run(
        "postgres",
        name=config.DB_CONTAINER_NAME,
        environment=[
            f"POSTGRES_USER={config.POSTGRES_USERNAME}",
            f"POSTGRES_PASSWORD={config.POSTGRES_PASSWORD}",
            f"POSTGRES_DB={config.POSTGRES_DATABASE}",
        ],
        detach=True,
        ports={"5432/tcp": config.CITROS_DATA_PORT},
    )
    # TODO: check container status...
    sleep(3)
    data_db_init(args, argv)
    print(
        f"[green]CITROS DB is running at: {config.CITROS_DATA_HOST}:{config.CITROS_DATA_PORT}"
    )


def data_db_init(args, argv):
    print(f"[green]CITROS Initializing DB...")
    _init_db(args.verbose, args.debug)
    print(
        f"[green]CITROS DB is running at: {config.CITROS_DATA_HOST}:{config.CITROS_DATA_PORT}"
    )


def data_db_status(args, argv):
    import docker

    try:
        client = docker.from_env()
    except Exception as e:
        print(
            "[red]Docker is not running. Please start docker and try again. exiting..."
        )
        if args.verbose:
            raise e
        return

    container = client.containers.get(config.DB_CONTAINER_NAME)
    # print(container)
    if container:
        print(
            f"[green]CITROS DB is running at: {container.attrs['NetworkSettings']['IPAddress']}:{container.attrs['NetworkSettings']['Ports']['5432/tcp'][0]['HostPort']}"
        )
    else:
        print(
            f"[red]CITROS DB is not running. Please run 'citros data db create' to create a new DB."
        )

    # console = Console()
    # with console.screen(hide_cursor=False) as screen:
    #     for line in container.stats(stream=True):
    #         stat = line.strip()
    #         stat = json.loads(stat)
    #         stat = json.dumps(stat, indent=4)
    #         # console.print(stat)
    #         screen.update(Panel(str(stat)))
    #         # inspect(stat)
    #         # sleep(5)
    #         #TODO: create status panel.


def data_db_stop(args, argv):
    import docker

    try:
        client = docker.from_env()
    except Exception as e:
        print(
            "[red]Docker is not running. Please start docker and try again. exiting..."
        )
        if args.verbose:
            raise e
        return

    try:
        container = client.containers.get(config.DB_CONTAINER_NAME)
        container.stop()
        print(f"[green]CITROS DB is stopped.")
    except docker.errors.NotFound:
        print(f"[green]CITROS DB is not running.")


def data_db_logs(args, argv):
    import docker

    try:
        client = docker.from_env()
    except Exception as e:
        print(
            "[red]Docker is not running. Please start docker and try again. exiting..."
        )
        if args.verbose:
            raise e
        return

    try:
        container = client.containers.get(config.DB_CONTAINER_NAME)
        console = Console()
        console.rule(
            f" Logs from CITROS database container: {config.DB_CONTAINER_NAME}"
        )
        for line in container.logs(stream=True, follow=False):
            print(line.decode("utf8").strip())
            # console.line(line.decode("utf8").strip())
            # console.log(line.decode("utf8").strip())

        console.rule()
    except docker.errors.NotFound:
        print(
            f"[red]CITROS DB is not running. Please run 'citros data db create' to create a new DB."
        )
        print(
            Panel.fit(
                Padding('You may run [green]"citros data db create" ', 1), title="help"
            )
        )


def data_db_remove(args, argv):
    import docker

    try:
        client = docker.from_env()
    except Exception as e:
        print(
            "[red]Docker is not running. Please start docker and try again. exiting..."
        )
        if args.verbose:
            raise e
        return

    container = client.containers.get(config.DB_CONTAINER_NAME)
    try:
        container.remove()
    except docker.errors.APIError as e:
        if e.status_code == 409:
            print("[red]CITROS DB is running. Please stop it before clearing.")
            print(
                Panel.fit(
                    Padding('You may run [green]"citros data db stop" ', 1),
                    title="help",
                )
            )
        else:
            raise e


############################# Service implementation ##############################
def service(args, argv):
    action = inquirer.select(
        raise_keyboard_interrupt=False,
        mandatory=False,
        keybindings=keybindings,
        message="Select Action:",
        choices=[
            Choice("start", name="▶  Start: starts CITROS API serivce."),
            Choice("stop", name="⏹  Stop: Stops the CITROS API service."),
            Choice("status", name="❓ Status: Show CITROS API status."),
            Separator(),
            Choice("exit", name="EXIT"),
        ],
        default="",
        border=True,
    ).execute()

    if action is None:
        exit_citros_cli()

    if action == "start":
        service_start(args, argv)
    elif action == "stop":
        service_stop(args, argv)
    elif action == "status":
        service_status(args, argv)
    elif action == "exit":
        exit_citros_cli()
    else:
        print("[red]Error: unknown action")


def service_start(args, argv):
    """
    :param args.dir
    :param args.debug:
    :param args.verbose:
    :param args.project_name:
    """
    from citros import data_access_service, NoDataFoundException

    host = getattr(args, "host", "0.0.0.0")
    port = getattr(args, "port", "8000")
    time = getattr(args, "time", False)

    root = Path(args.dir).expanduser().resolve() / ".citros/data"
    print(
        Panel.fit(
            f"""started at [green]http://{host}:{port}[/green].
API: open [green]http://{host}:{port}/redoc[/green] for documantation
Listening on: [green]{str(root)}""",
            title="[green]CITROS service",
        )
    )
    try:
        # TODO[important]: make async
        data_access_service(
            str(root),
            time=time,
            host=host,
            port=int(port),
            debug=args.debug,
            verbose=args.verbose,
        )
    except NoDataFoundException:
        print(
            f'[red] "{Path(args.dir).expanduser().resolve()}" has not been initialized. cant run "citros data service" on non initialized directory.'
        )
        return


# TODO: implement
def service_stop(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


# TODO[important]: implement  after making this sevice async. return status of service.
def service_status(args, argv):
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


############################# Reports implementation ##############################


def report(args, argv):
    # inspect(args)
    # print(
    #     Panel(
    #         Markdown(
    #             open(
    #                 importlib_resources.files(f"data.doc.cli").joinpath("report.md"),
    #                 "r",
    #             ).read()
    #         ),
    #         title="citros report",
    #         subtitle=f"[{citros_version}]",
    #     )
    # )
    # action
    action = inquirer.select(
        raise_keyboard_interrupt=False,
        mandatory=False,
        keybindings=keybindings,
        message="Select Action:",
        choices=[
            Choice("list", name="*️⃣  List: reports list "),
            Choice("generate", name="⚡ Generate: new report"),
            Choice("validate", name="❓ Validate: report integrity"),
            Separator(),
            Choice("exit", name="EXIT"),
        ],
        default="",
        border=True,
    ).execute()

    if action is None:
        exit_citros_cli()

    if action == "list":
        report_list(args, argv)
    elif action == "generate":
        report_generate(args, argv)
    elif action == "validate":
        report_validate(args, argv)
    elif action == "exit":
        exit_citros_cli()
    else:
        print("[red]Error: unknown action")


def report_list(args, argv):
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
        flat_repo = citros.get_reports_flat()
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    table = Table(
        title=f"Reports from: [blue]{citros.root_citros / 'reports'}", box=box.SQUARE
    )
    table.add_column("date", style="cyan", no_wrap=False)
    # table.add_column("started_at", style="cyan", no_wrap=True)
    # table.add_column("finished_at", style="cyan", no_wrap=True)
    table.add_column("name", style="magenta", justify="left")
    table.add_column("Versions", justify="left", style="green")
    table.add_column("message", style="magenta", justify="left")
    table.add_column("progress", justify="right", style="green")
    table.add_column("status", style="magenta", justify="left")
    table.add_column(
        "path", style="cyan", justify="left", no_wrap=False, overflow="fold"
    )
    _name = None
    for flat in flat_repo:
        table.add_row(
            flat["started_at"],
            # flat["finished_at"],
            None if flat["name"] == _name else flat["name"],
            flat["version"],
            flat["message"],
            str(flat["progress"]),
            flat["status"],
            str(flat["path"]).removeprefix(os.getcwd()).removeprefix("/"),
            # f"[link={flat['path']}]path[/link]",
        )
        _name = flat["name"]

    console = Console()
    console.print(table)


def report_generate(args, argv):
    """
    Handle the 'generate' command for Citros report.

    :param args.execute: Flag to indicate execution of notebooks.
    :param args.render: Flag to indicate rendering of notebooks to PDF.
    :param args.sign: Flag to indicate signing of PDFs.
    :param args.key_path: Path to the private key file for signing PDFs.
    :param args.notebooks: List of paths to Jupyter notebooks.
    :param args.style_path: Path to the CSS style file, if any.

    :param args.dir
    :param args.name
    :param args.message
    :param args.output: Path to the output folder for generated files.
    """
    # inspect(args)
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f"[red]Error:[/red] {Path(args.dir).expanduser().resolve()} has not been initialized with citros."
        )
        print(Panel.fit(Padding("You may run [green]citros init ", 1), title="help"))
        return

    simulation, batch_name, version = choose_batch(
        citros.root_citros,
        simulation_name=getattr(args, "simulation", None),
        batch_name=getattr(args, "batch", None),
        version=getattr(args, "version", None),
    )

    if not hasattr(args, "name") or args.name is None:
        report_name = Prompt.ask("Please name this report", default="citros")
    else:
        report_name = args.name

    if not hasattr(args, "message") or args.message is None:
        report_message = Prompt.ask(
            "Enter a message for this report",
            default="This is a default report message from citros",
        )
    else:
        report_message = args.message

    if not hasattr(args, "output"):
        output = None
    else:
        output = args.output

    if not hasattr(args, "notebooks"):
        notebook_list = []
        for notebook in glob.glob(f"{os.getcwd()}/**/*.ipynb", recursive=True):
            notebook_list.append(notebook.removeprefix(os.getcwd()).removeprefix("/"))

        notebooks = inquirer.select(
            raise_keyboard_interrupt=False,
            mandatory=False,
            keybindings=keybindings,
            message="Select notebook:",
            choices=notebook_list,
            border=True,
            multiselect=True,
            instruction="Use [space] to select notebooks, [enter] to confirm selection.",
            mandatory_message="Please select at least one notebook",
        ).execute()

        # print(f"chosen_notebooks: {notebooks}")

    else:
        notebooks = args.notebooks

    if not hasattr(args, "sign"):
        sign = False  # TODO: change to True
    else:
        sign = args.sign

    batch = citros.get_batch(
        simulation,
        batch_name,
        version,
    )
    # inspect(batch)
    # print(
    #     f"report_name={report_name}, report_message={report_message}, output={output}, sign={sign}"
    # )
    report = Report(
        name=report_name,
        message=report_message,
        citros=citros,
        output=output,
        batch=batch,
        notebooks=notebooks,
        sign=sign,
        log=citros.log,
        debug=args.debug,
        verbose=args.verbose,
    )

    # if args.debug:
    console = Console()
    console.rule(f"command")
    print(
        f'[white]citros report generate {"--output " + str(output) if output is not None else ""} {" ".join(["-nb " + str(nb) for nb in notebooks])} --dir {args.dir} --simulation {simulation} --batch {batch_name} --version {version} --name {report_name} --message "{report_message}" {"--sign" if sign else ""} '
    )
    console.rule(f"")

    # Execute notebooks
    print("[green]Executing notebook...")
    folder = report.run()

    print(f'[green]Report generation completed at [blue]"{folder}"')


def report_validate(args, argv):
    """
    Handle the 'validate' command for Citros report.

    :param args.check: Flag to indicate verification of PDF signatures.
    :param args.public_key_path: Path to the public key file for verification.
    :param args.pdfs: List of paths to PDF files to be verified.
    """

    # Extract arguments
    check_flag = args.check
    public_key_path = args.public_key_path
    pdf_paths = args.pdfs

    # Validate arguments
    if not check_flag:
        print("Error: Check flag is not set.")
        return

    if not public_key_path:
        print("Error: Missing public key for verification.")
        return

    if not pdf_paths:
        print("Error: No PDF paths provided for verification.")
        return

    # Verify PDF signatures
    for pdf_path in pdf_paths:
        if Report.validate(pdf_path, public_key_path):
            print(f"The contents of {pdf_path} are intact.")
        else:
            print(f"Warning: The contents of {pdf_path} may have been altered.")

    print("PDF verification completed.")


## Utils


def choose_batch(
    root_citros: Path, simulation_name=None, batch_name=None, version=None
):
    data_root = root_citros / "data"

    chosen_simulation = simulation_name
    if chosen_simulation is None:
        simulations = []
        for simulation_path in glob.glob(f"{data_root}/[!_]*/"):
            simulation = simulation_path.removesuffix("/").split("/")[-1]
            simulations.append(simulation)
        # print(f"simulations: {simulations}")
        if simulations == []:
            print("[yellow]Warning: No simulations found. Please create a simulation.")
            exit_citros_cli()
        chosen_simulation = inquirer.select(
            raise_keyboard_interrupt=False,
            mandatory=False,
            keybindings=keybindings,
            message="Select Simulation:",
            choices=simulations,
            default="",
            border=True,
        ).execute()
    # print(f"chosen_simulation: {chosen_simulation}")
    if chosen_simulation is None:
        exit_citros_cli()

    chosen_batch = batch_name
    if chosen_batch is None:
        batch_list = []
        for batch_path in glob.glob(f"{data_root}/{chosen_simulation}/[!_]*/"):
            batch_name = batch_path.removesuffix("/").split("/")[-1]
            batch_list.append(batch_name)

        chosen_batch = inquirer.select(
            raise_keyboard_interrupt=False,
            mandatory=False,
            keybindings=keybindings,
            message="Select Batch:",
            choices=batch_list,
            default="",
            border=True,
        ).execute()

    # print(f"chosen_batch: {chosen_batch}")

    if chosen_batch is None:
        exit_citros_cli()

    chosen_version = version
    if chosen_version is None:
        version_list = []
        for version_path in glob.glob(
            f"{data_root}/{chosen_simulation}/{chosen_batch}/[!_]*/"
        ):
            version = version_path.removesuffix("/").split("/")[-1]
            version_list.append(version)
        if version_list == []:
            print("[yellow]Warning: No versions found for this batch.")
            exit_citros_cli()
        chosen_version = inquirer.select(
            raise_keyboard_interrupt=False,
            mandatory=False,
            keybindings=keybindings,
            message="Select Version:",
            choices=version_list,
            default="",
            border=True,
        ).execute()

    if chosen_version is None:
        exit_citros_cli()

    return chosen_simulation, chosen_batch, chosen_version



import os
import time
import psycopg2
import importlib_resources

from jinja2 import Template
from pathlib import Path
from .logger import get_logger, shutdown_log


class CitrosDB:
    def __init__(
        self,
        db_user="citros",
        db_password="password",
        db_host="localhost",
        db_port="5454",
        db_name="citros",
        organization_name="citros",
        log=None,
        verbose=False,
        debug=False,
    ) -> None:
        self.db_user = db_user
        self.db_password = db_password
        self.db_host = db_host
        self.db_port = db_port
        self.db_name = db_name
        self.organization_name = organization_name
        self.verbose = verbose
        self.debug = debug
        self._init_log(Path.cwd())

    ###################
    ##### private #####
    ###################
    def _init_log(self, root=None, log=None):
        self.log = log
        if self.log is None:
            log_dir = root / "logs"

            if not log_dir.exists():
                Path.home().joinpath(".citros/logs").mkdir(parents=True, exist_ok=True)
                log_dir = Path.home().joinpath(".citros/logs")

            self.log = get_logger(
                __name__,
                log_level=os.environ.get("LOGLEVEL", "DEBUG" if self.debug else "INFO"),
                log_file=str(log_dir / "citros.log"),
                verbose=self.verbose,
            )

    def init_db(self):
        """
        Initialize the database by creating the organization's database and executing SQL scripts.

        Args:
            organization_name (str): The name of the organization.
            db_user (str): The username for the database connection.
            db_password (str): The password for the database connection.
            db_host (str): The host address of the database server.
            db_port (int): The port number of the database server.
            db_name (str): The name of the database.

        Raises:
            Exception: If failed to render the SQL template.

        Returns:
            None
        """

        connection = self.connect()
        cursor = connection.cursor()
        connection.autocommit = True

        # Define variables for rendering the template
        context = {
            "ORGANIZATION_NAME": self.organization_name,
            "USER_NAME": self.db_user,
            "USER_PASSWORD": self.db_password,
        }
        # Render the template with the provided context
        with open(
            importlib_resources.files(f"data.sql").joinpath(
                "templates/create_db.sql.j2"
            ),
            "r",
        ) as file_:
            template = Template(file_.read())
            rendered_sql = template.render(context)

        if rendered_sql is None:
            raise Exception("Failed to render sql template")
        # cursor.execute(f'CREATE DATABASE "{organization_name}";')
        try:
            self.log.debug(rendered_sql)
            cursor.execute(rendered_sql)
        except psycopg2.DatabaseError as e:
            self.log.error(f"Database already exist ... log: {e}")

        connection.close()

    def connect(self):
        """
        Establish a connection to a PostgreSQL database and return a connection object and cursor.

        This method tries to establish a connection to the PostgreSQL database using the
        setup parameters provided in `pg_setup`. If the connection is successful, it returns
        the connection object and a cursor to interact with the database. In case of an
        OperationalError like "too many clients", it retries connecting.

        Parameters:
        -----------
        pg_setup : dict
            A dictionary containing PostgreSQL setup parameters, including:
                - username: Database username
                - password: Database password
                - host: Database host address
                - port: Database port number
                - db_name: Database name

        Returns:
        --------
        connection : psycopg2.extensions.connection or None
            The connection object for the database, or None if the connection fails.

        Raises:
        -------
        psycopg2.OperationalError:
            An error from the psycopg2 driver when unable to connect to the database.

        Examples:
        ---------
        >>> connect_to_pg({'username': 'user', 'password': 'pass', 'host': 'localhost', 'port': '5432', 'db_name': 'db'})
        # (psycopg2.extensions.connection, psycopg2.extensions.cursor)

        """
        connection = None
        retries = 0
        sleep_durations = [2**x for x in range(10)]
        while not connection:
            try:
                connection = psycopg2.connect(
                    user=self.db_user,
                    password=self.db_password,
                    host=self.db_host,
                    port=self.db_port,
                    database=self.db_name,
                )
                self.log.info("Successfully connected to Postgres Database")

                return connection

            except psycopg2.OperationalError as e:
                if "too many clients" in str(e).lower():
                    time.sleep(sleep_durations[retries])
                    retries = retries + 1 if retries < 9 else 9
                    self.log.warning(e)
                else:
                    self.log.error(e)
                    break

    def create_table(self, connection, schema_name, table_name):
        from jinja2 import Environment, FileSystemLoader

        if connection is None:
            self.log.error("No connection to database, cant create table.")
            return
        self.log.debug(f"creating table: {schema_name}.{table_name}")

        cursor = connection.cursor()

        context = {
            "ORGANIZATION_NAME": self.organization_name,
            "SCHEMA_NAME": schema_name,
            "TABLE_NAME": table_name,
        }

        with open(
            importlib_resources.files(f"data.sql").joinpath(
                "templates/create_table.sql.j2"
            ),
            "r",
        ) as file_:
            template = Template(file_.read())
            # Render the template with the provided context
            rendered_sql = template.render(context)

        # TODO: query check if table exists. if not, create it.
        try:
            cursor.execute(rendered_sql)
        except Exception as ex:
            self.log.error(ex)
            return

    def drop_table(self, connection, schema_name, table_name):
        from jinja2 import Environment, FileSystemLoader

        self.log.debug(f"dropping table: {schema_name}.{table_name}")

        cursor = connection.cursor()

        context = {
            "SCHEMA_NAME": schema_name,
            "TABLE_NAME": table_name,
        }

        with open(
            importlib_resources.files(f"data.sql").joinpath(
                "templates/drop_table.sql.j2"
            ),
            "r",
        ) as file_:
            template = Template(file_.read())
            # Render the template with the provided context
            rendered_sql = template.render(context)

        # TODO: query check if table exists. if not, create it.
        try:
            cursor.execute(rendered_sql)
            connection.commit()
            self.log.debug(rendered_sql)
        except Exception as ex:
            self.log.error(ex)



import os
import json
import logging
from pathlib import Path
from rich.logging import RichHandler
from rich import print, inspect, print_json

# import logger
from .logger import get_logger, shutdown_log


class CitrosException(Exception):
    def __init__(self, message="Simulation exception."):
        super().__init__(message)


class CitrosNotFoundException(CitrosException):
    def __init__(self, message="No batch found."):
        super().__init__(message)


class FileNotFoundException(CitrosException):
    def __init__(self, message="No batch found."):
        super().__init__(message)


class NoValidException(CitrosException):
    def __init__(self, message="No batch found."):
        super().__init__(message)


class CitrosObj:
    """Object representing .citros/*.json filew."""

    #####################
    ##### overrides #####
    #####################
    def __init__(
        self,
        name: str,
        root=None,
        new=False,
        log=None,
        citros=None,
        verbose=False,
        debug=False,
        level=0,
    ):
        """CITROS object is a folder and a set of json file that is located in the .citros folder

        Args:
            root (str): the root path to the `parent` of .citros folder
            name (str): the name of the pbject file
            new (bool, optional): will create new object if true. Defaults to False.
            log (logging, optional): log. Defaults to None.
            citros (CitrosObj, optional): the main citros object that represent "project.json" file. Defaults to None.

        Raises:
            CitrosException: _description_
        """
        self.verbose = verbose
        self.debug = debug
        self.level = level

        self.name = name
        if name.endswith(".json"):
            self.name = name[:-5]
        self.file = self.name + ".json"

        if root is None:
            raise CitrosException(f'root cant be "{root}"')
        if name is None:
            raise CitrosException(f'name cant be "{name}"')

        self.new = new

        # self.root = the path to the parent of .citros object.
        # self.root = Path.cwd() if root is None else Path(root).expanduser().resolve()
        # self.root = self._find_citros_in_ancestors(new=new)
        if root is None:
            self.root = (
                Path.cwd() if root is None else Path(root).expanduser().resolve()
            )
        else:
            self.root = Path(root)

        # path to .citros folder
        self.root_citros = self.root / ".citros"

        self._init_log(log)
        self.log.debug(f"Citros root={self.root}")

        self.log.debug(
            f"{'   '*level}{self.__class__.__name__}.init(name={self.name}, new={new})",
        )

        # holds tha main citros object.
        self.citros = citros
        if citros is None:
            self.citros = self

        self.data = {}
        if self.new:
            if self.path().exists():
                self._load()
            else:
                self._new()
        else:
            self._load()

        self._validate()

    def __str__(self):
        # print_json(data=self.data)
        return json.dumps(self.data, indent=4)

    def __getitem__(self, key):
        """get element from object

        Args:
            key (str): the element key

        Returns:
            str: the element value
        """
        return self.data[key]

    def get(self, key, default=None):
        """get element from object

        Args:
            key (str): the element key

        Returns:
            str: the element value
        """
        return self.data.get(key, default)

    def __setitem__(self, key, newvalue):
        self.data[key] = newvalue
        self.save()

    ###################
    ##### private #####
    ###################
    def _init_log(self, log=None):
        self.log = log
        if self.log is None:
            log_dir = self.root_citros / "logs"
            if self.new:
                log_dir.mkdir(parents=True, exist_ok=True)
            else:
                if not log_dir.exists():
                    Path.home().joinpath(".citros/logs").mkdir(
                        parents=True, exist_ok=True
                    )
                    log_dir = Path.home().joinpath(".citros/logs")

            self.log = get_logger(
                __name__,
                log_level=os.environ.get("LOGLEVEL", "DEBUG" if self.debug else "INFO"),
                log_file=str(log_dir / "citros.log"),
                verbose=self.verbose,
            )

    def _validate(self) -> bool:
        """Validate .json file."""
        raise NoValidException()

    def _save(self):
        """Save .json file."""

        # print("_save self.path()", self.path())
        with open(self.path(), "w") as file:
            json.dump(self.data, file, indent=4, sort_keys=True)

    def _create(self):
        """_summary_
            this should be idempotent function:
                can be applied multiple times without changing the result
            if file exists -> check and append values if needded
            else -> create new file

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError

    def _load(self):
        if not Path(self.root_citros).exists():
            raise CitrosNotFoundException(f'"{self.root_citros}" does not exist.')

        if not Path(self.path()).exists():
            raise FileNotFoundException(f'"{self.path()}" does not exist.')

        """Load .json file."""
        with open(self.path(), "r") as file:
            self.data = json.load(file)

    def _new(self):
        """initialize citros object Create .json file."""
        raise NotImplementedError

    def _find_citros_in_ancestors(self, proj_dir=None, new=False):
        current_dir = (
            Path.cwd() if proj_dir is None else Path(proj_dir).expanduser().resolve()
        )

        return current_dir
        # print("current_dir", current_dir)
        # Ensure we don't go into an infinite loop at the root directory
        while current_dir != current_dir.parent:
            citros_dir = current_dir / ".citros"
            # print("citros_dir", citros_dir)
            if citros_dir.exists():
                return citros_dir.expanduser().resolve()
            current_dir = current_dir.parent

        return None

    ##################
    ##### public #####
    ##################
    def path(self):
        """return the full path to the current main file.

        default to ".citros/project.json"

        Returns:
            str: the full path to the current main file.
        """
        return self.root_citros / self.file



# ==============================================
#  ██████╗██╗████████╗██████╗  ██████╗ ███████╗
# ██╔════╝██║╚══██╔══╝██╔══██╗██╔═══██╗██╔════╝
# ██║     ██║   ██║   ██████╔╝██║   ██║███████╗
# ██║     ██║   ██║   ██╔══██╗██║   ██║╚════██║
# ╚██████╗██║   ██║   ██║  ██║╚██████╔╝███████║
#  ╚═════╝╚═╝   ╚═╝   ╚═╝  ╚═╝ ╚═════╝ ╚══════╝
# ==============================================

from .database import CitrosDB
from .citros import Citros
from .citros_obj import (
    CitrosException,
    CitrosNotFoundException,
    FileNotFoundException,
    NoValidException,
)
from .utils import str_to_bool, suppress_ros_lan_traffic
from .batch import Batch
from .batch_uploader import NoConnectionToCITROSDBException
from .logger import get_logger, shutdown_log
from .service import data_access_service, NoDataFoundException

# reporting
from .report import Report, NoNotebookFoundException

__all__ = [
    Citros,
    CitrosException,
    CitrosNotFoundException,
    FileNotFoundException,
    NoValidException,
    get_logger,
    shutdown_log,
    CitrosDB,
    data_access_service,
    NoDataFoundException,
    Batch,
    str_to_bool,
    suppress_ros_lan_traffic,
    Report,
    NoNotebookFoundException,
]



import uuid
from pathlib import Path


class citros_batch:
    def __init__(self, citros):
        self.citros = citros
        self.log = citros.log

    def generate_batch_id(self):
        # generate a random UUID
        return str(uuid.uuid4())

    def get_batch(self, batch_run_id, simulation_name=None):
        batch = None

        if self.citros.CITROS_ENVIRONMENT != "CLUSTER":
            assert (
                simulation_name is not None
            ), "Simulation name cannot be None while running locally."

            user_commit, user_branch = self.citros.get_git_info(
                self.citros.USER_PROJ_DIR
            )
            citros_commit, citros_branch = self.citros.get_git_info(
                self.citros.CITROS_REPO_DIR
            )

            sim_info = self.citros.get_simulation_info(simulation_name)

            batch = {
                "id": str(self.citros._batch_id),
                "simulation": simulation_name,
                "message": self.citros._batch_message,
                "name": self.citros._batch_name,
                "citrosCommit": citros_commit,
                "citrosBranch": citros_branch,
                "userCommit": user_commit,
                "userBranch": user_branch,
                "cpu": sim_info["CPU"],
                "gpu": sim_info["GPU"],
                "memory": sim_info["MEM"],
                "storageType": sim_info["storage_type"],
                "timeout": sim_info["timeout"],
                "metadata": None,
            }
        else:
            query = """
            query getData($batchRunId: UUID!) {
            batchRun(id: $batchRunId) {
                    id
                    simulation                
                    message
                    name
                    citrosCommit
                    citrosBranch
                    userCommit
                    userBranch
                    cpu
                    gpu
                    memory
                    storageType,
                    timeout
                    metadata
                }
            }
            """
            batch = self.citros.gql_execute(
                query, variable_values={"batchRunId": batch_run_id}
            )["batchRun"]

        return batch

    def create_batch(
        self,
        batch_id,
        repo_id,
        simulation_name,
        gpu,
        cpu,
        memory,
        storage_type,
        completions,
        user_commit,
        user_branch,
        citros_commit,
        citros_branch,
        latest_tag,
        timeout,
        name="",
        message="",
        parallelism=1,
        metadata={},
    ):
        query = """
        mutation createBatchRun($batch_run: CreateBatchRunInput!) {
            createBatchRun(input: $batch_run) {
            batchRun {
                id
            }
            }
        }
        """
        result = self.citros.gql_execute(
            query,
            variable_values={
                "batch_run": {
                    "batchRun": {
                        "id": batch_id,
                        "repoId": repo_id,
                        "simulation": simulation_name,
                        "gpu": gpu,
                        "cpu": cpu,
                        "memory": int(memory),
                        "storageType": storage_type,
                        "name": name,
                        "message": message,
                        "completions": int(completions),
                        "parallelism": parallelism,
                        "citrosBranch": citros_branch,
                        "citrosCommit": citros_commit,
                        "userBranch": user_branch,
                        "userCommit": user_commit,
                        "tag": latest_tag,
                        "metadata": metadata,
                        "timeout": timeout,
                    }
                }
            },
        )

        return result["createBatchRun"]["batchRun"]["id"]



from io import StringIO
import traceback
import datetime 
import sqlite3
import json
import os


CHUNK_SIZE = 50 * 1024 * 1024

###############
##### PG ######
############### 
class BagReaderSQL():
    def read_messages(self, input_bag, simulation_run_id):
        print(" + read_messages:")
        from rosidl_runtime_py.convert import get_message_slot_types, message_to_yaml, message_to_csv, message_to_ordereddict
        from rosidl_runtime_py.utilities import get_message
        from rclpy.serialization import deserialize_message
        
        self.conn = sqlite3.connect(input_bag)
        self.cursor = self.conn.cursor()

        topics_data = self.cursor.execute("SELECT id, name, type FROM topics").fetchall()
        self.topics = {name_of:{'type':type_of, 'id':id_of, 'message':get_message(type_of) } for id_of,name_of,type_of in topics_data}
        
        buffer = StringIO()
        size = 0
        for topic_name in self.topics.keys():
            if topic_name in ["/rosout", "/client_count", "/connected_clients"]:
                continue
            rows = self.cursor.execute(f"select id, timestamp, data from messages where topic_id = {self.topics[topic_name]['id']}").fetchall()            
            rid = 0
            for id, timestamp, data in rows:
                d_data = deserialize_message(data, self.topics[topic_name]["message"])
                msg_dict = message_to_ordereddict(d_data)
                json_data = json.dumps(msg_dict)
                
                row = chr(0x1E).join([f"{simulation_run_id}", f"{rid}", f"{timestamp}", f"{topic_name}", f"{self.topics[topic_name]['type']}", json_data])
                rid = rid + 1
                bytes_wrote = buffer.write(row + '\n')
                size = size + bytes_wrote
                # 10MB chunks max
                if size >= CHUNK_SIZE:                    
                    yield buffer
                    buffer, size = StringIO(), 0
        # return the rest of the file.
        yield buffer



from io import StringIO
import traceback
import datetime 
import os

import json
import yaml

class BagReaderMcap():
    def read_messages(self, input_bag, simulation_run_id):             
        from rosidl_runtime_py.convert import get_message_slot_types, message_to_yaml, message_to_csv, message_to_ordereddict
        from rosidl_runtime_py.utilities import get_message
        from rclpy.serialization import deserialize_message
        import rosbag2_py

        reader = rosbag2_py.SequentialReader()
        reader.open(
            rosbag2_py.StorageOptions(uri=input_bag, storage_id="mcap"),
            rosbag2_py.ConverterOptions(
                input_serialization_format="cdr", output_serialization_format="cdr"
            ),
        )

        topic_types = reader.get_all_topics_and_types()

        def typename(topic_name):
            for topic_type in topic_types:
                if topic_type.name == topic_name:
                    return topic_type.type
            raise ValueError(f"topic {topic_name} not in bag")
        
        topic_counters = {}
        buffer = StringIO()    
        
        while reader.has_next():  
                 
            topic, data, timestamp = reader.read_next()
            
            msg_type = get_message(typename(topic))
            
            msg = deserialize_message(data, msg_type)
            
            yaml_msg = message_to_yaml(msg)
            
            data = yaml.load(yaml_msg, Loader=yaml.FullLoader)
            
            json_data = json.dumps(data)
            
            # msg_dict = message_to_ordereddict(msg)
            # print("json:", topic, msg_type, "\n",json_data)            
            
            id = topic_counters.get(topic, -1) + 1
            topic_counters[topic] = id
            # TODO:[ ] return ros message id?
            row = chr(0x1E).join([f"{sid}", f"{id}", f"{timestamp}", f"{topic}", f"{msg_type}", f"{json_data}\n"])            
            buffer.write(row)       
                                
        del reader
        return buffer



import sys
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install
from bin import __version__ as citros_version

install()


# citros report
def parser_report_generate(subparser, epilog=None):
    description_path = "report/generate.md"
    help = "citros report generate section"
    parser = subparser.add_parser(
        "generate",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")

    # namings
    parser.add_argument(
        "-n",
        "--name",
        nargs="?",
        default=None,
        help="a name for the run",
        required=True,
    )
    parser.add_argument(
        "-m",
        "--message",
        nargs="?",
        default=None,
        help="a message for the run",
        required=True,
    )

    # data source
    parser.add_argument(
        "-s", "--simulation", nargs="?", help="simulation name", required=True
    )
    parser.add_argument("-b", "--batch", nargs="?", help="batch name", required=True)
    parser.add_argument("-ver", "--version", nargs="?", default=-1, help="version")

    # notebooks
    parser.add_argument(
        "-nb", "--notebooks", nargs="+", help="Paths to Jupyter notebooks"
    )
    parser.add_argument(
        "-style", "--style-path", help="Path to CSS style file", default=None
    )
    parser.add_argument("-output", "--output", help="Path to output folder")

    # sign
    parser.add_argument("--sign", action="store_true", help="Sign PDFs")
    parser.add_argument(
        "-key",
        "--key-path",
        help="Path to private key file for signing",
    )
    # debug
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=report_generate)

    return parser


# citros report list
def parser_report_list(subparser, epilog=None):
    description_path = "report/list.md"
    help = "citros report list section"
    parser = subparser.add_parser(
        "list",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")

    # debug
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=report_list)

    return parser


# citros report validate
def parser_report_validate(subparser, epilog=None):
    description_path = "report/validate.md"
    help = "citros report validate section"
    parser = subparser.add_parser(
        "validate",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-c", "--check", action="store_true", help="Verify PDF signatures"
    )
    # parser.add_argument("paths", nargs='*', help="Path to the public key file for verification, followed by paths to PDF files to be verified")
    parser.add_argument(
        "-key",
        "--public-key-path",
        help="Path to public key file for verification",
        required=True,
    )
    parser.add_argument(
        "-pdfs", nargs="+", help="Paths to PDF files to be verified", required=True
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=report_validate)
    return parser


# citros report
def parser_report(subparser, epilog=None):
    description_path = "report.md"
    help = "report section"
    parser = subparser.add_parser(
        "report",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="citros report",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-n", "--name", default=None, help="name of report")
    parser.add_argument("-m", "--match", default=None, help="match report pattern")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )

    parser.set_defaults(func=report)

    subparser = parser.add_subparsers(dest="type")
    # report run/list
    parser_report_generate(subparser, epilog=epilog)
    parser_report_list(subparser, epilog=epilog)
    parser_report_validate(subparser, epilog=epilog)

    return parser



import json
import uuid

from pathlib import Path

from .citros_obj import CitrosObj

from .utils import validate_file


class Settings(CitrosObj):
    """Object representing .citros/settings.json file."""

    ###################
    ##### private #####
    ###################

    # overriding
    def _validate(self):
        """Validate simulation.json file."""

        success = validate_file(self.path(), "schema_settings.json", self.log)

        return success

    # overriding
    def _new(self):
        path = self.path()

        # avoid overwrite
        if path.exists():
            self._load()
            # return

        Path(self.root).mkdir(parents=True, exist_ok=True)

        default = {
            "name": "default_settings",
            "force_message": "True",
            "force_batch_name": "True",
        }
        self.data = {**default, **self.data}

        self._save()
